"""Build of python module wrapping geolib.c, as defined by CFFI lib.

"""

import os
from cffi import FFI

ffi = FFI()
SOURCE = os.path.join(os.path.split(__file__)[0], 'geolib.c')

ffi.cdef("""
void exactinit(void);
double orient2dfast(double *pa, double *pb, double *pc);
double orient2d(double *pa, double *pb, double *pc);
double orient3dfast(double *pa, double *pb, double *pc, double *pd);
double orient3d(double *pa, double *pb, double *pc, double *pd);
double incirclefast(double *pa, double *pb, double *pc, double *pd);
double incircle(double *pa, double *pb, double *pc, double *pd);
double inspherefast(double *pa, double *pb, double *pc, double *pd, double *pe);
double insphere(double *pa, double *pb, double *pc, double *pd, double *pe);
bool pred_incirclefast(double *pa, double *pb, double *pc, double *pd);
bool pred_incircle(double *pa, double *pb, double *pc, double *pd);
bool pred_incirclefast_strict(double *pa, double *pb, double *pc, double *pd);
bool pred_incircle_strict(double *pa, double *pb, double *pc, double *pd);
bool pred_inspherefast(double *pa, double *pb, double *pc, double *pd, double *pe);
bool pred_insphere(double *pa, double *pb, double *pc, double *pd, double *pe);
bool pred_inspherefast_strict(double *pa, double *pb, double *pc, double *pd, double *pe);
bool pred_insphere_strict(double *pa, double *pb, double *pc, double *pd, double *pe);
""")

with open(SOURCE) as sfd:
    ffi.set_source("gerobust._wrapped_geolib", sfd.read(),
                   extra_compile_args=['-O3', '-frounding-math', '-fsignaling-nans'],
                   libraries=['c'])

if __name__ == "__main__":
    ffi.compile()



"""Wrapper around geolib.c file.

"""


import os
import ctypes

from gerobust._wrapped_geolib import lib as GEOLIB


GEOLIB.exactinit()


def orientation_fast(pa:(float, float), pb:(float, float), pc:(float, float)) -> float:
    """Approximate 2D orientation test. Nonrobust.

    Return a positive value if the points pa, pb, and pc occur
    in counterclockwise order; a negative value if they occur
    in clockwise order; and zero if they are collinear.

    The result is also a rough approximation of twice the signed
    area of the triangle defined by the three points.

    Do not use exact arithmetic, therefore is quicker than orientation.

    """
    return GEOLIB.orient2dfast(pa, pb, pc)

def orientation(pa:(float, float), pb:(float, float), pc:(float, float)) -> float:
    """Adaptive exact 2D orientation test. Robust.

    Return a positive value if the points pa, pb, and pc occur
    in counterclockwise order; a negative value if they occur
    in clockwise order; and zero if they are collinear.

    The result is also a rough approximation of twice the signed
    area of the triangle defined by the three points.

    Use exact arithmetic to ensure a correct answer.
    The result returned is the determinant of a matrix.
    This determinant is here computed adaptively, in the sense that exact
    arithmetic is used only to the degree it is needed to ensure that the
    returned value has the correct sign.

    Hence, this function is usually quite fast,
    but will run more slowly when the input points are collinear or nearly so.

    """
    return GEOLIB.orient2d(pa, pb, pc)


def orientation_3d_fast(pa:(float, float), pb:(float, float),
                        pc:(float, float), pd:(float, float)) -> float:
    """Approximate 3D orientation test. Nonrobust.

    Return a positive value if the point pd lies below the
    plane passing through pa, pb, and pc; "below" is defined so
    that pa, pb, and pc appear in counterclockwise order when
    viewed from above the plane. Returns a negative value if
    pd lies above the plane. Returns zero if the points are
    coplanar. The result is also a rough approximation of six
    times the signed volume of the tetrahedron defined by the
    four points.

    Do not use exact arithmetic, therefore is quicker than orientation_3d.

    """
    return GEOLIB.orient3dfast(pa, pb,
                               pc, pd)

def orientation_3d(pa:(float, float), pb:(float, float),
                   pc:(float, float), pd:(float, float)) -> float:
    """Adaptive exact 3D orientation test. Robust.

    Return a positive value if the point pd lies below the
    plane passing through pa, pb, and pc; "below" is defined so
    that pa, pb, and pc appear in counterclockwise order when
    viewed from above the plane. Returns a negative value if
    pd lies above the plane. Returns zero if the points are
    coplanar. The result is also a rough approximation of six
    times the signed volume of the tetrahedron defined by the
    four points.

    Use exact arithmetic to ensure a correct answer.
    The result returned is the determinant of a matrix.
    This determinant is computed adaptively, in the sense that exact
    arithmetic is used only to the degree it is needed to ensure that the
    returned value has the correct sign. Hence, this function is usually quite
    fast, but will run more slowly when the input points are coplanar or
    nearly so.

    """
    return GEOLIB.orient3d(pa, pb,
                           pc, pd)


def incirclefast(pa:(float, float), pb:(float, float),
                 pc:(float, float), pd:(float, float)) -> float:
    """Approximate 2D incircle test. Nonrobust.

    Return a positive value if the point pd lies inside the
    circle passing through pa, pb, and pc; a negative value if
    it lies outside; and zero if the four points are cocircular.
    The points pa, pb, and pc must be in counterclockwise
    order, or the sign of the result will be reversed.

    Do not use exact arithmetic, therefore is quicker than incircle.

    """
    return GEOLIB.incirclefast(pa, pb,
                               pc, pd)

def incircle(pa:(float, float), pb:(float, float),
             pc:(float, float), pd:(float, float)) -> float:
    """Adaptive exact 2D incircle test. Robust.

    Return a positive value if the point pd lies inside the
    circle passing through pa, pb, and pc; a negative value if
    it lies outside; and zero if the four points are cocircular.
    The points pa, pb, and pc must be in counterclockwise
    order, or the sign of the result will be reversed.

    Use exact arithmetic to ensure a correct answer.
    The result returned is the determinant of a matrix.
    This determinant is computed adaptively,
    in the sense that exact arithmetic is used only to the degree it is needed
    to ensure that the returned value has the correct sign.
    Hence, this function is usually quite fast, but will run more slowly
    when the input points are cocircular or nearly so.

    """
    return GEOLIB.incircle(pa, pb, pc, pd)


def inspherefast(pa:(float, float), pb:(float, float), pc:(float, float),
                 pd:(float, float), pe:(float, float)) -> float:
    """Approximate 3D insphere test. Nonrobust.

    Return a positive value if the point pe lies inside the
    sphere passing through pa, pb, pc, and pd; a negative value
    if it lies outside; and zero if the five points are
    cospherical. The points pa, pb, pc, and pd must be ordered
    so that they have a positive orientation,
    or the sign of the result will be reversed.

    Do not use exact arithmetic, therefore is quicker than insphere.

    """
    return GEOLIB.inspherefast(pa, pb, pc, pd, pe)

def insphere(pa:(float, float), pb:(float, float), pc:(float, float),
             pd:(float, float), pe:(float, float)) -> float:
    """Adaptive exact 3D insphere test. Robust.

    Return a positive value if the point pe lies inside the
    sphere passing through pa, pb, pc, and pd; a negative value
    if it lies outside; and zero if the five points are
    cospherical. The points pa, pb, pc, and pd must be ordered
    so that they have a positive orientation (as defined by
    orient3d()), or the sign of the result will be reversed.

    Use exact arithmetic to ensure a correct answer.
    The result returned is the determinant of a matrix.
    This determinant is computed adaptively,
    in the sense that exact arithmetic is used only to the degree it is needed
    to ensure that the returned value has the correct sign.
    Hence, this function is usually quite fast, but will run more slowly
    when the input points are cospherical or nearly so.

    """
    return GEOLIB.insphere(pa, pb, pc, pd, pe)


# shortcuts
orientation.fast = orientation_fast
orientation_3d.fast = orientation_3d_fast
incircle.fast = incirclefast
insphere.fast = inspherefast



from gerobust.wrapper import (orientation, orientation_fast,
                              orientation_3d, orientation_3d_fast,
                              incircle, incirclefast,
                              insphere, inspherefast)
from gerobust.predicates import (clockwise, clockwise_fast,
                                 counter_clockwise, counter_clockwise_fast)
from gerobust import predicates as pred





import itertools
from collections import deque

import pytest
from gerobust import orientation
from gerobust import counter_clockwise, clockwise


def sliding(it:iter, size:int) -> iter:
    it = iter(it)
    acc = deque(itertools.islice(it, 0, size), maxlen=size)
    yield tuple(acc)
    for item in it:
        acc.append(item)
        yield tuple(acc)

def test_sliding():
    assert ((1, 2, 3), (2, 3, 4)) == tuple(sliding((1, 2, 3, 4), size=3))

def test_list_usage():
    assert orientation([0, 0], [0, 50], [50, 50])

def test_clockwise():
    points = (0, 0), (0, 50), (50, 50)
    assert orientation(*points) == -orientation(*reversed(points))
    assert     clockwise(*points)
    assert not counter_clockwise(*points)
    assert not clockwise(*reversed(points))
    assert     counter_clockwise(*reversed(points))

def test_clockwise_2():
    points = (0, 0), (50, 50), (100, 0)
    assert orientation(*points) == -orientation(*reversed(points))
    assert     clockwise(*points)
    assert not counter_clockwise(*points)
    assert not clockwise(*reversed(points))
    assert     counter_clockwise(*reversed(points))

def test_counter_clockwise():
    points = (0, 0), (100, 0), (50, 50)
    assert orientation(*points) == -orientation(*reversed(points))
    assert not clockwise(*points)
    assert     counter_clockwise(*points)
    assert     clockwise(*reversed(points))
    assert not counter_clockwise(*reversed(points))


def test_orientation():
    points = (5,0), (6,4), (4,5)
    assert counter_clockwise(*points)
    assert clockwise(*reversed(points))


def test_orientation_2():
    triangles = (
        ((3, 4), (5, 6), (9, 5), True),
        ((5, 6), (9, 5), (12, 8), False),
        ((9, 5), (12, 8), (5, 11), False),
        ((12, 8), (5, 11), (3, 4), False),
    )
    for *points, is_clockwise in triangles:
         # = triangle
        assert clockwise(*points) is is_clockwise
        assert counter_clockwise(*points) is not is_clockwise
        assert clockwise(*reversed(points)) is not is_clockwise
        assert counter_clockwise(*reversed(points)) is is_clockwise



import os
from setuptools   import setup, find_packages
from pip.req      import parse_requirements
from pip.download import PipSession


# import the data inside package/info.py file, without trigger the
#  importing of the whole package.
#  equivalent to the python2 execfile routine.
SRC_DIR = 'gerobust'
INFO_FILE = SRC_DIR + '/info.py'
with open(INFO_FILE) as fd:
    code = compile(fd.read(), INFO_FILE, 'exec')
    local_vars = {}
    exec(code, {}, local_vars)  # don't use global vars, save local_vars
    __pkg_name__ = local_vars['PACKAGE_NAME']  # save the interesting data
    __version__ = local_vars['PACKAGE_VERSION']


# access to the file at the package top level (like README)
def path_to(filename):
    return os.path.join(os.path.dirname(__file__), filename)

# parse_requirements() returns generator of pip.req.InstallRequirement objects
install_reqs = parse_requirements(path_to('requirements.txt'),
                                  session=PipSession())
reqs = [str(ir.req) for ir in install_reqs]


setup(
    name=__pkg_name__,
    version=__version__,
    packages=find_packages(),
    include_package_data=True,  # read the MANIFEST.in file
    install_requires=reqs,
    setup_requires=["cffi>=1.0.0"],
    zip_safe=False,

    cffi_modules=["./gerobust/build_geolib.py:ffi"],

    author='lucas bourneuf',
    author_email='lucas.bourneuf@openmailbox.org',
    description='Extension in C for incircles tests (2D/3D)',
    long_description=open(path_to('README.mkd')).read(),
    keywords='C function robust geometry',
    url='https://github.com/aluriak/gerobust',

    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Natural Language :: English',
        'Programming Language :: C',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
